/**   @file  oom.c
 *   @brief
 *  @author  François Cerbelle (Fanfan), francois@cerbelle.net
 *
 *  @internal
 *       Created:  21/06/2024
 *      Revision:  none
 * Last modified:  2024-07-26 00:57
 *      Compiler:  gcc
 *  Organization:  Cerbelle.net
 *     Copyright:  Copyright (c) 2024, François Cerbelle
 *
 *  This source code is released for free distribution under the terms of the
 *  GNU General Public License as published by the Free Software Foundation.
 */

#include "oom.h"                                /* OOM simulation */

#ifndef __GNUC__
# ifndef __asm__
#  define __asm__ asm
# endif
# ifndef __sync_synchronize
#  define __sync_synchronize void
# endif
#endif

#include "debug/assert.h"                       /* libdebug's Assertions */
#include <sys/resource.h>                       /* setrlimit */
#include <errno.h>                              /* errno */
#include <string.h>                             /* strerror() */
#include <stdio.h>                              /* printf */
#include <malloc.h>                             /* malloc_trim() */
#include <stdlib.h>                             /* abort() */
#include <alloca.h>                             /* alloca() */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <unistd.h>                             /* sysconf() */

/** Maximum number of fragmented blocks to allocate
 *
 * I never reached more than 350 on my systems.
 * The value needs to be less than UINT_MAX
 * */
#define RAMBLOCKS_MAX 1000

/** RAM eating pointer
 *
 * Used to store the huge memory block used to fill the available memory */
static void* _oomblocks[RAMBLOCKS_MAX] = {0};

/** getrlimit helper with abort on fail */
static int checked_getrlimit(int resource, struct rlimit *rlim)
{
    /* Get current limit values */
    if (getrlimit(resource, rlim) != 0) {
        /* Can occur, thus not ignored, but impossible to trigger for gcov/lcov */
        fprintf (stderr,"%s:%d getrlimit() failed with errno=%d %s\n",
                 __FILE__,__LINE__, errno,strerror(errno));
        abort();
    }
    return 0;
}

/** Disabled fill function
 *
 * Used as oomtest_fill when disabled. Thus, the same test can be surrounded by
 * fill/free, but the actual RAM pressure will or will not be activated.
 *
 * \sa oomtest_fill
 * \sa oomtest_free
 * */
static size_t oomtest_fill_preinit(const size_t minHeap, const size_t minStack)
{
    (void)minHeap;
    (void)minStack;
    ASSERT(NULL==_oomblocks[0]);
    return 0;
}

/* Documented in header file */
size_t (*oomtest_fill)(const size_t minHeap, const size_t minStack)=oomtest_fill_preinit;

/** Disabled free function
 *
 * Used as oomtest_free when disabled. Thus, the same test can be surrounded by
 * fill/free, but the actual RAM pressure will or will not be activated.
 *
 * \sa oomtest_free
 * \sa oomtest_fill
 * */
static void oomtest_free_preinit()
{
    ASSERT(NULL==_oomblocks[0]);
}

/* Documented in header file */
void (*oomtest_free)()=oomtest_free_preinit;

/** Disabled oomtest_enable function
 *
 * Used as oomtest_enable before oomtest_config was invoked, which should never
 * happen and will fail to detect mistakes in the test code.
 *
 * \sa oomtest_enable
 * \sa oomtest_disable
 * */
static size_t oomtest_enable_preinit(const size_t softlimit)
{
    (void)softlimit;

    /* Should not be called without configuration */
    fprintf(stderr,"%s:%d oomtest_enable called without oomtest_config before\n",__FILE__,__LINE__);
    abort();
}

/* Documented in header file */
size_t (*oomtest_enable)(const size_t softlimit)=oomtest_enable_preinit;

/** Disabled oomtest_disable function
 *
 * Used as oomtest_disable before oomtest_config was invoked, which should never
 * happen and will fail to detect mistakes in the test code.
 *
 * \sa oomtest_enable
 * \sa oomtest_disable
 * */
static size_t oomtest_disable_preinit()
{
    /* Should not be called before configuration */
    fprintf(stderr,"%s:%d oomtest_disable called without oomtest_config before\n",__FILE__,__LINE__);
    abort();
}

/* Documented in header file */
size_t (*oomtest_disable)()=oomtest_disable_preinit;

/** Finds biggest available RAM block and allocate it
 *
 * This functions tries to find and allocate the biggest available memory block
 * by dichotomy (Olog2 complexity) between 0 and rlim_cur soft limit. It returns
 * the pointer.
 *
 * \param [in,out] The allocated pointer if successfull, NULL otherwise
 *
 * \return The size of the allocated block, 0 if not block was allocated.
 *
 * */
static size_t oomtest_getbiggestblock(void** p_ramblock)
{
    /* This function could receive an optimized "max" value and return the final
     * "max" value to the caller. So, it could be used as the next invocation
     * starting "max" value. This would save few loop iterations, at the cost of
     * extra stack usage, I chose to not pass this value as a parameter to avoid
     * consuming stack.
     * The function starts to search between 0..rlim_cur which has very little
     * impact given the Olog2 complexity. It could be optimized and start to
     * search between 0..sysconf(AVPHYSPAGE*PAGESIZE), assuming that sysconf is
     * faster than few loop iterations. */

    /* Use static to avoid stack allocation/free */
    static size_t max,cur;
    static struct rlimit limit;

    ASSERT(NULL!=p_ramblock);
    ASSERT(NULL==*p_ramblock);

    /* Get the current limits */
    checked_getrlimit(RLIMIT_AS, &limit);
    max = limit.rlim_cur;

    /* Restart the whole process if cur can not be allocated at the end */
    while ((max>0)&&(NULL==*p_ramblock)) {
        static size_t min;
        /* Iterate quickly (Olog2) to converge to the biggest available RAM block */
        min = 0;
        while (max>min) {
            cur = min+(max-min)/2; /* To avoid overflow */
            if (NULL==(*p_ramblock = malloc(cur))) {
                max = cur;
            } else {
                min = cur+1;
                free(*p_ramblock);
            }
        }
        cur -= 1;
        *p_ramblock=malloc(cur);
    }

    return cur;
}

/** Enabled fill function
 *
 * Used as oomtest_fill when enabled.
 *
 * \sa oomtest_fill
 * \sa oomtest_free
 * */
static size_t oomtest_fill_postinit(const size_t minHeap, const size_t minStack)
{
    unsigned int l_numblock;
    size_t l_sum;
    void* volatile l_reservedheap;
    void* l_reservedstack;

    /* Probably a mistake in the test code, do not accept despite we could */
    if (NULL!=_oomblocks[0]) {
        /* Dirty hack to free some RAM and allow abort to SIGABRT */
        free(_oomblocks[0]);
        fprintf (stderr,"%s:%d oomblocks are already allocated\n", __FILE__,__LINE__);
        abort();
    }

    /* Reserve/Protect stack bytes which will be auto freed at return */
    /* A failure means a stackoverflow, which is not recoverable and will abort
     * the process anyway. */
    if (0<minStack)
        l_reservedstack=alloca(minStack);
    (void)l_reservedstack;

    /* Reserve/Protect heap bytes which will be released before return */
    l_reservedheap=NULL;
    if (0<minHeap)
        if (NULL==(l_reservedheap=malloc(minHeap))) {
            /* This will fail if minHeap is higher than rlimit */
            fprintf(stderr,"%s:%d Failed to reserve minheap bytes\n",__FILE__,__LINE__);
            abort();
        }

    /* Find and allocate the biggest available RAM blocks until no mre RAM
     * available or all _oomblocks are allocated */
    l_numblock=0;
    l_sum = 0;
    l_sum += oomtest_getbiggestblock(&(_oomblocks[l_numblock++]));
    while ((NULL!=_oomblocks[l_numblock-1])&&((RAMBLOCKS_MAX-1)>l_numblock))
        l_sum += oomtest_getbiggestblock(&(_oomblocks[l_numblock++]));
    /* Either already NULL, which stopped the while loop or reached the last
     * slot in the table, which stopped the while loop and the slot has to be
     * set to NULL */
    _oomblocks[l_numblock]=NULL;

    /* There can be less than 4 bytes available at this point !!! */

    /* Make the protected heap bytes available again */
    if (NULL!=l_reservedheap)
        free(l_reservedheap);

    return l_sum;
}

/** Enabled free function
 *
 * Used as oomtest_free when enabled
 *
 * \sa oomtest_free
 * \sa oomtest_fill
 * */
static void oomtest_free_postinit()
{
    unsigned int l_i;

    /* Despite we could manage, abort to help detecting mistakes in test code */
    if (NULL==_oomblocks[0]) {
        fprintf(stderr,"%s:%d no blocks to free in oomtest_free.\n",__FILE__,__LINE__);
        abort();
    }

    /* Actually free allocated blocks and set their pointer to NULL */
    l_i = 0;
    while ((l_i<(RAMBLOCKS_MAX-1))&&(_oomblocks[l_i])) {
        free(_oomblocks[l_i]);
        _oomblocks[l_i] = NULL;
        l_i+=1;
    }
    /* These barrier protections are probably useless */
    malloc_trim(0);
    /* SW barrier (compiler only) */
    __asm__ volatile("": : :"memory");
    /* HW barrier (CPU instruction) */
    __sync_synchronize();
}

/** Enabled oomtest_enable function
 *
 * Used as oomtest_enable after oomtest_config was invoked
 *
 * \sa oomtest_enable
 * \sa oomtest_disable
 * */
static size_t oomtest_enable_postinit(const size_t softlimit)
{
    struct rlimit limit;
    size_t l_softlimit = softlimit;

    /* Probably a bug in oomtest */
    ASSERT(NULL==_oomblocks[0]);

    /* Get current limit values */
    checked_getrlimit(RLIMIT_AS, &limit);

    /* Check the requested value */
    if (0==l_softlimit) {
        /* Defaults to available physical RAM if requested 0 */
        l_softlimit = limit.rlim_max;
    };

    /* Soft limit available RAM */
    limit.rlim_cur = l_softlimit;

    /* Abort if setrlimit fails to avoid RAM bombing */
    if (setrlimit(RLIMIT_AS, &limit) != 0) {
        fprintf (stderr,"%s:%d setrlimit(cur=%lu, max=%lu) with errno=%d %s\n",
                 __FILE__,__LINE__,
                 (unsigned long)limit.rlim_cur, (unsigned long)limit.rlim_max,
                 errno,strerror(errno));
        abort();
    }

    /* Activate fill/free functions */
    oomtest_fill = oomtest_fill_postinit;
    oomtest_free = oomtest_free_postinit;

    /* Return the actual current soft limit */
    return limit.rlim_cur;
}

/** Enabled oomtest_disable function
 *
 * Used as oomtest_disable after oomtest_config was invoked
 *
 * \sa oomtest_enable
 * \sa oomtest_disable
 * */
static size_t oomtest_disable_postinit()
{
    struct rlimit limit;

    if (NULL!=_oomblocks[0]) {
        fprintf(stderr,"%s:%d RAM still allocated while calling oomtest_disable\n",__FILE__,__LINE__);
        abort();
    }

    /* Do not allow calling disable if not enabled to detect test mistakes */
    if ((oomtest_fill!=oomtest_fill_postinit)||(oomtest_free!=oomtest_free_postinit)) {
        fprintf(stderr,"%s:%d Impossible to disable oomtest if not previously enabled\n",__FILE__,__LINE__);
        abort();
    }

    /* Reset the soft limit to hard limit */
    oomtest_enable_postinit(0);

    /* Get current limit values */
    checked_getrlimit(RLIMIT_AS, &limit);

    /* Restore disabled functors */
    oomtest_fill = oomtest_fill_preinit;
    oomtest_free = oomtest_free_preinit;

    /* Return the actual current soft limit, which is equal to hard limit */
    return limit.rlim_cur;
}

/* Documented in header file */
size_t oomtest_config(const size_t hardlimit)
{
    struct rlimit limit;
    size_t l_avail;
    size_t l_hardlimit = hardlimit;

    /* Probably a test implementation mistake */
    if (NULL!=_oomblocks[0]) {
        fprintf(stderr,"%s:%d Calling oomtest_config with allocated RAM blocks is not allowed.\n",
                __FILE__,__LINE__);
        abort();
    }

    /* Find *installed* physical RAM, 0 in case of failure */
    l_avail = (sysconf(_SC_PHYS_PAGES) * sysconf(_SC_PAGESIZE));

    /* Get current limit values */
    checked_getrlimit(RLIMIT_AS, &limit);
    if (0==l_hardlimit) {
        /* Defaults to available physical RAM or already set rlimit
         * if 0 requested */
        l_hardlimit=(limit.rlim_max<l_avail?limit.rlim_max:l_avail);
    } else if (l_hardlimit>l_avail) {
        /* Fails if request is over available physical RAM to avoir swapping */
        fprintf(stderr,"%s:%d Requesing a limit %lu bigger than *installed* RAM %lu is not allowed.\n",
                __FILE__,__LINE__,
                (unsigned long)l_hardlimit,(unsigned long)l_avail);
        abort();
    }

    /* Hard limit available RAM to hardlimit globally with no way back */
    limit.rlim_cur = l_hardlimit;
    limit.rlim_max = l_hardlimit;

    /* Abort if setrlimit fails to avoid RAM bombing */
    if (setrlimit(RLIMIT_AS, &limit) != 0) {
        fprintf (stderr,"%s:%d setrlimit(cur=%lu, max=%lu) with errno=%d %s\n",
                 __FILE__,__LINE__, (unsigned long)limit.rlim_cur,
                 (unsigned long)limit.rlim_max, errno,strerror(errno));
        /* Get current limit values */
        checked_getrlimit(RLIMIT_AS, &limit);
        fprintf (stderr,"%s:%d getrlimit() is cur=%lu, max=%lu\n",
                 __FILE__,__LINE__, (unsigned long)limit.rlim_cur,
                 (unsigned long)limit.rlim_max);
        abort();
    }

    /* Activate enable/disable functions */
    oomtest_enable = oomtest_enable_postinit;
    oomtest_disable = oomtest_disable_postinit;

    /* Return the configured limit hard=soft */
    return limit.rlim_max;
}

unsigned char oomtest_enabled()
{
    return (oomtest_fill==oomtest_fill_postinit?1:0);
}
